

places <- []


class MagicPlace
{
  sensor=null
  optional=true
  activated=false


  function destroy()
  {
    if (sensor) { sensor.destroy(); sensor=null }
  }


  function beforeErosion() {}
  function  afterErosion() {}
  function beforeObjects() {}
}


class MagicCircle extends MagicPlace
{
  cx=0
  cz=0
  rad=0

  action=null


  constructor(x, z, r, a)
  {
    cx=x; cz=z; rad=r
    action=a
  }


  function intersects(p)
  {
    if (!p) return false;

    if (p instanceof MagicCircle)
    {
      local dx=wrap(p.cx-cx)
      local dz=wrap(p.cz-cz)
      return dx*dx+dz*dz < p.rad*p.rad+2*p.rad*rad+rad*rad
    }
    else
      return p.intersects(this);
  }


  function beforeErosion()
  {
    flatten_circle(cx, cz, rad*1.5, get_average_circle_height(cx, cz, rad), 0.8, 2)
  }


  function afterErosion()
  {
    local y=get_average_circle_height(cx, cz, rad)
    flatten_circle(cx, cz, rad*1.5, y, 0.5, 2)
  }


  function beforeObjects()
  {
    local y=get_land_height(cx, cz)

    if (frnd()<0.3)
    {
      create_stone_circle(cx, cz, rad)
      if (frnd()<0.5) create_stone_circle(cx, cz, rad*0.8)
    }
    else
    {
      create_menhir_circle(cx, cz, rad)
    }

    if (!sensor && !in_main_menu)
    {
      sensor=Sensor(vec4f(cx, y, cz, 1), rad)
      sensor.setOnEvent(function()
      {
        activated=true
        sensor.destroy(); sensor=null
        if (action) action()
      }.bindenv(this))
    }
  }
}


monster_spawn_enabled<-true


function spawn_birds()
{
  if (!monster_spawn_enabled) return;
  local cpos=vec4f(cx, get_land_height(cx, cz)-5, cz, 1)
  for (local n=30; n; --n)
  {
    local dir=make_dir_xz(frnd())
    BirdMonster(cpos+dir*(rad*1.2), dir)
  }
}


function spawn_wyverns()
{
  if (!monster_spawn_enabled) return;
  local cpos=vec4f(cx, get_land_height(cx, cz)-10, cz, 1)
  for (local n=3; n; --n)
  {
    local dir=make_dir_xz(frnd())
    WyvernMonster(cpos+dir*(rad*1.5))
  }
}


function spawn_mana()
{
  local cpos=vec4f(cx, get_land_height(cx, cz)-10, cz, 1)
  for (local n=30; n; --n)
  {
    local dir=make_dir_xz(frnd())
    ManaBall(cpos+dir*(rad*1.2), rndf(0.1, 0.2), vec4f(0, rndf(15, 30), 0, 0))
  }
}


function spawn_portal()
{
  WorldPortal(vec4f(cx, get_land_height(cx, cz), cz, 1))
  action=null
}


function create_stone_circle(cx, cz, rad)
{
  local cpos=vec4f(cx, 0, cz, 1)
  local num=(rad*(1.5+srnd()*0.1)).tointeger()
  local ofs=frnd()
  for (local i=0; i<num; ++i)
  {
    local p=cpos+make_dir_xz((i+srnd()*0.2)/num+ofs)*(rad*(1+srnd()*0.04))
    Stone1StaticObject(getx(p), getz(p))
  }
}


function create_menhir_circle(cx, cz, rad)
{
  local cpos=vec4f(cx, 0, cz, 1)
  local num=(rad*(0.5+srnd()*0.1)).tointeger()
  if (num<3) num=3
  local ofs=frnd()
  for (local i=0; i<num; ++i)
  {
    local p=cpos+make_dir_xz((i+srnd()*0.1)/num+ofs)*(rad*(1+srnd()*0.04));
    (frnd()<0.5?Menhir1StaticObject:Menhir2StaticObject)(getx(p), getz(p))
  }
}


function new_magic_place()
{
  local action=null, opt=false
  local r=frnd()
  if (r<0.1) { action=spawn_mana; opt=true }
  else if (r<0.55) action=spawn_wyverns
  else action=spawn_birds

  local p=MagicCircle(hmrnd(), hmrnd(), rndf(5, 20), action)
  p.optional=opt
  return p;
}


function check_portal()
{
  if (!portal_place || !portal_place.action) return;

  foreach (p in places)
    if (!p.optional && !p.activated) return;

  if (are_any_monsters_alive()) return;

  portal_place.action=spawn_portal
  portal_place.action()
}


function check_conditions()
{
  check_portal()
}


function create_magic_places()
{
  local p=MagicCircle(hmrnd(), hmrnd(), rndf(10, 15), check_portal)
  places.append(p)
  portal_place<-p

  local iter=1000
  for (local num=irnd()%5+5; num>0; )
  {
    if (--iter<0) break; //== fatal error?

    local np=new_magic_place()

    local ok=true
    foreach (p in places)
      if (np.intersects(p)) { ok=false; break; }

    if (!ok) { np.destroy(); continue; }

    places.append(np)
    --num
  }
}
